@@ -82,6 +82,7 @@ typedef struct {
 	gchar         *hash_name;
 	gint           in_fd;
 	gint           out_fd;
+	gint           tty_fd;
 	GIOChannel    *error_channel;
 	pid_t          ssh_pid;
 
@@ -1047,7 +1048,7 @@ typedef enum {
 } SFTPClientVendor;
 
 static SFTPClientVendor
-get_sftp_client_vendor ()
+get_sftp_client_vendor (void)
 {
 	char *ssh_stderr;
 	char *args[3];
@@ -1071,6 +1072,7 @@ get_sftp_client_vendor ()
 		else
 			res = SFTP_VENDOR_INVALID;
 	}
+	g_free (ssh_stderr);
 	g_free (args[0]);
 	g_free (args[1]);
 
@@ -1083,8 +1085,8 @@ static GnomeVFSResult
 sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri) 
 {
 	GnomeVFSResult  res;
-	GIOChannel     *error_channel;
-	GIOChannel     *tty_channel;
+	GIOChannel     *error_channel = NULL;
+	GIOChannel     *tty_channel = NULL;
 	int             in_fd, out_fd, err_fd, tty_fd;
 	pid_t           ssh_pid;
 	const gchar    *user_name;
@@ -1275,16 +1277,21 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 			if (g_str_has_suffix (buffer, "password: ") ||
 			    g_str_has_suffix (buffer, "Password: ") ||
 			    g_str_has_prefix (buffer, "Enter passphrase for key")) {
-				if (!done_auth && invoke_fill_auth (uri, buffer, &password) && password != NULL) {
-					g_io_channel_write_chars (prompt_channel, password, -1, &len, NULL);
-					g_io_channel_write_chars (prompt_channel, "\n", 1, &len, NULL);
-					g_io_channel_flush (prompt_channel, NULL);
+				if (!done_auth && gnome_vfs_uri_get_password (uri) != NULL) {
+					password = g_strdup (gnome_vfs_uri_get_password (uri));
+					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
+					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
+				} else if (!done_auth && invoke_fill_auth (uri, buffer, &password) && password != NULL) {
+					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
+					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
 				} else if (invoke_full_auth (uri, done_auth, buffer, &password, &keyring, 
 							     &user, &object, &authtype, &save_password) && password != NULL) {
 					full_auth = TRUE;
-					g_io_channel_write_chars (prompt_channel, password, -1, &len, NULL);
-					g_io_channel_write_chars (prompt_channel, "\n", 1, &len, NULL);
-					g_io_channel_flush (prompt_channel, NULL);
+					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
+					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
 				} else {
 					res = GNOME_VFS_ERROR_ACCESS_DENIED;
 					goto bail;
@@ -1356,19 +1363,19 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 				
 				if (invoked) {
 					if (out_args.answer == 0) {
-						g_io_channel_write_chars (prompt_channel, "yes\n", -1, &len, NULL);
+						g_io_channel_write_chars (tty_channel, "yes\n", -1, &len, NULL);
 					} else {
-						g_io_channel_write_chars (prompt_channel, "no\n", -1, &len, NULL);
+						g_io_channel_write_chars (tty_channel, "no\n", -1, &len, NULL);
 						g_free (in_args.primary_message);
 						g_free (in_args.secondary_message);
 						res = GNOME_VFS_ERROR_ACCESS_DENIED;
 						goto bail;
 					}
-					g_io_channel_flush (prompt_channel, NULL);
+					g_io_channel_flush (tty_channel, NULL);
 					buffer[0]='\0';
 				} else {
-					g_io_channel_write_chars (prompt_channel, "no\n", -1, &len, NULL);
-					g_io_channel_flush (prompt_channel, NULL);
+					g_io_channel_write_chars (tty_channel, "no\n", -1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
 					buffer[0]='\0';
 					g_free (in_args.primary_message);
 					g_free (in_args.secondary_message);
@@ -1409,7 +1416,9 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 		(*connection)->ref_count = 1;
 		(*connection)->in_fd = in_fd;
 		(*connection)->out_fd = out_fd;
+		(*connection)->tty_fd = tty_fd;
 		(*connection)->error_channel = error_channel;
+		g_io_channel_ref (error_channel);
 		(*connection)->ssh_pid = ssh_pid;
 		(*connection)->version = buffer_read_gint32 (&msg);
 		(*connection)->mutex = g_mutex_new ();
@@ -1432,11 +1441,18 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 	g_free (user);
 	g_free (object);
 	g_free (authtype);
-
+	if (error_channel != NULL) {
+		g_io_channel_unref (error_channel);
+	}
+	if (tty_channel != NULL) {
+		g_io_channel_unref (tty_channel);
+	}
 	if (res != GNOME_VFS_OK) {
 		close (in_fd);
 		close (out_fd);
 		close (err_fd);
+		if (tty_fd != -1)
+			close (tty_fd);
 		*connection = NULL;
 
 		/* TODO: Do we leak error_channel and connection? */
@@ -1469,7 +1485,10 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		user_name = g_get_user_name ();
 
 	if (host_name == NULL)
-		return GNOME_VFS_ERROR_HOST_NOT_FOUND;
+	{
+		res = GNOME_VFS_ERROR_HOST_NOT_FOUND;
+		goto bail;
+	}
 
 	hash_name = g_strconcat (user_name, "@", host_name, NULL);
 
@@ -1485,7 +1504,12 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = sftp_connect (connection, uri);
 
 		if (res == GNOME_VFS_OK) {
-			g_return_val_if_fail (*connection != NULL, GNOME_VFS_ERROR_INTERNAL);
+			if (*connection == NULL)
+			{
+				res = GNOME_VFS_ERROR_INTERNAL;
+				g_free (hash_name);
+				goto bail;
+			}
 			g_mutex_lock ((*connection)->mutex);
 			(*connection)->hash_name = hash_name;
 			g_hash_table_insert (sftp_connection_table, hash_name, *connection);
@@ -1517,6 +1541,7 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = GNOME_VFS_OK;
 	}
 
+ bail:
 	G_UNLOCK (sftp_connection_table);
 
 	return res;
@@ -1529,6 +1554,8 @@ sftp_connection_close (SftpConnection *conn)
 
 	close (conn->in_fd);
 	close (conn->out_fd);
+	if (conn->tty_fd != -1)
+		close (conn->tty_fd);
 	g_source_remove (conn->event_id);
 	g_io_channel_shutdown (conn->error_channel, FALSE, NULL);
 	g_io_channel_unref (conn->error_channel);
@@ -2566,9 +2593,9 @@ do_read_directory (GnomeVFSMethod       *method,
 			while (handle->info_write_ptr + count > handle->info_alloc) {
 				handle->info_alloc *= 2;
 				handle->info = g_renew (GnomeVFSFileInfo, handle->info, handle->info_alloc);
-				bzero (&(handle->info[handle->info_write_ptr]),
-				       sizeof (GnomeVFSFileInfo) *
-				       (handle->info_alloc - handle->info_write_ptr));
+				memset (&(handle->info[handle->info_write_ptr]), 0,
+					sizeof (GnomeVFSFileInfo) *
+					(handle->info_alloc - handle->info_write_ptr));
 			}
 		}
 
@@ -2646,7 +2673,7 @@ do_make_directory (GnomeVFSMethod  *method,
 	id = sftp_connection_get_id (conn);
 
 	path = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (uri), NULL);
-	bzero (&info, sizeof (GnomeVFSFileInfo));
+	memset (&info, 0, sizeof (GnomeVFSFileInfo));
 	iobuf_send_string_request_with_file_info (conn->out_fd, id, SSH2_FXP_MKDIR,
 						  path, strlen (path), &info,
 						  GNOME_VFS_SET_FILE_INFO_NONE);
